package de.hft.stuttgart.citydoctor2.gui;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ValidationConfiguration;
import de.hft.stuttgart.citydoctor2.check.error.SchematronError;
import de.hft.stuttgart.citydoctor2.datastructure.*;
import de.hft.stuttgart.citydoctor2.exceptions.CityDoctorWriteException;
import de.hft.stuttgart.citydoctor2.gui.table.ErrorStat;
import de.hft.stuttgart.citydoctor2.gui.tree.*;
import de.hft.stuttgart.citydoctor2.mapper.citygml3.GMLValidationHandler;
import de.hft.stuttgart.citydoctor2.parser.*;
import de.hft.stuttgart.citydoctor2.utils.Localization;
import javafx.application.Platform;
import javafx.scene.chart.XYChart.Data;
import javafx.scene.chart.XYChart.Series;
import javafx.scene.control.ButtonType;
import javafx.scene.control.Dialog;
import javafx.scene.control.TreeItem;
import javafx.scene.control.TreeView;
import javafx.stage.FileChooser;
import javafx.stage.FileChooser.ExtensionFilter;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.citygml4j.core.model.core.CityModel;
import org.xml.sax.SAXParseException;

import java.io.File;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;

public class CityDoctorController {

    private static final int MAX_FEATURES_PER_CHUNK = 200;

    private static final Logger logger = LogManager.getLogger(CityDoctorController.class);

    private MainWindow mainWindow;

    private CityDoctorModel model;
    private ParserConfiguration currentConfig;
    private String sourceFile;

    private Checker currentChecker;

    private HighlightController highlightController;
    private Renderer renderer;
    private BoundingBox originBB = null;

    private AtomicInteger buildingChunkNr = new AtomicInteger(0);
    private AtomicInteger vegetationChunkNr = new AtomicInteger(0);
    private AtomicInteger transportationChunkNr = new AtomicInteger(0);
    private AtomicInteger tunnelChunkNr = new AtomicInteger(0);
    private AtomicInteger bridgeChunkNr = new AtomicInteger(0);
    private AtomicInteger waterChunkNr = new AtomicInteger(0);
    private AtomicInteger landChunkNr = new AtomicInteger(0);
    private AtomicInteger cityFurnitureChunkNr = new AtomicInteger(0);
    private AtomicInteger otherObjectsChunkNr = new AtomicInteger(0);

    public CityDoctorController(MainWindow mainWindow, HighlightController highlightController, Renderer renderer) {
        this.mainWindow = mainWindow;
        this.highlightController = highlightController;
        this.renderer = renderer;
    }

    public void loadCityGml(String path, int numberOfRoundingPlaces, ProgressListener l, boolean useValidation)
            throws CityGmlParseException, InvalidGmlFileException {
        loadCityGml(path, numberOfRoundingPlaces, l, useValidation, false);
    }

    public void loadCityGml(String path, int numberOfRoundingPlaces, ProgressListener l, boolean useValidation,
                            boolean lowMemory) throws CityGmlParseException, InvalidGmlFileException {
        try {
            Platform.runLater(() -> {
                mainWindow.getOpenBtn().setDisable(true);
                mainWindow.getWriteReportButton().setDisable(true);
                mainWindow.getMeshGroup().getChildren().clear();

                clearTrees();
                mainWindow.resetSearchBar();
                mainWindow.resetFilterComboBox();
            });
            currentChecker = null;
            currentConfig = new ParserConfiguration(numberOfRoundingPlaces, useValidation, lowMemory);
            GMLValidationHandler handler = null;
            List<String> validationIssues = new ArrayList<>();
            if (useValidation) {
                handler = new GMLValidationHandler() {
                    @Override
                    public void error(SAXParseException exception) {
                        if (exception.getLineNumber() >= 0) {
                            String s = "In line " + exception.getLineNumber() + ":";
                            validationIssues.add(s);
                        }
                        validationIssues.add(exception.getMessage());
                    }

                    @Override
                    public void warning(SAXParseException exception) {
                        error(exception);
                    }

                    @Override
                    public void fatalError(SAXParseException exception) {
                        error(exception);
                    }
                };
            }
            model = CityGmlParser.parseCityGmlFile(path, currentConfig, l, handler);
            if (!validationIssues.isEmpty()) {
                StringJoiner sj = new StringJoiner("\n");
                validationIssues.stream().forEach(sj::add);
                throw new InvalidGmlFileException(sj.toString());
            }
            mainWindow.getClickHandler().setConfig(currentConfig);
            sourceFile = path;
            renderer.reset();
            Platform.runLater(() -> {
                mainWindow.addFileNameToTitle(path);
                mainWindow.getCheckButton().setDisable(false);
                mainWindow.getLod1Btn().setDisable(false);
                mainWindow.getLod2Btn().setDisable(false);
                mainWindow.getLod3Btn().setDisable(false);
                mainWindow.getLod4Btn().setDisable(false);
                mainWindow.getWorldBtn().setDisable(false);
                mainWindow.getSaveBtn().setDisable(false);
                mainWindow.getResetCameraBtn().setDisable(false);
                mainWindow.getHideRoofBtn().setDisable(false);
                mainWindow.getNorthArrow().setVisible(true);
                mainWindow.alignNorthArrow();
                setupFeatureTabs();
                buildTrees();
            });
        } finally {
            Platform.runLater(() -> mainWindow.getOpenBtn().setDisable(false));
        }
    }

    private void setupFeatureTabs() {
        mainWindow.setDisableOfFeatureTab(FeatureType.BUILDING, model.getBuildings().isEmpty());
        mainWindow.setDisableOfFeatureTab(FeatureType.VEGETATION, model.getVegetation().isEmpty());
        mainWindow.setDisableOfFeatureTab(FeatureType.TRANSPORTATION, model.getTransportation().isEmpty());
        mainWindow.setDisableOfFeatureTab(FeatureType.TUNNEL, model.getTunnels().isEmpty());
        mainWindow.setDisableOfFeatureTab(FeatureType.BRIDGE, model.getBridges().isEmpty());
        mainWindow.setDisableOfFeatureTab(FeatureType.WATER, model.getWater().isEmpty());
        mainWindow.setDisableOfFeatureTab(FeatureType.LAND, model.getLand().isEmpty());
        mainWindow.setDisableOfFeatureTab(FeatureType.CITY_FURNITURE, model.getCityFurniture().isEmpty());
        mainWindow.setDisableOfFeatureTab(FeatureType.GENERIC_CITY_OBJECT, model.getGenericCityObjects().isEmpty());
    }

    public void buildTrees() {
        resetFeatureChunks();
        buildBuildings(model);
        buildVegetation(model.getVegetation());
        buildTransportation(model.getTransportation());
        buildBridges(model);
        buildWater(model);
        buildLand(model);
        buildCityFurniture(model);
        buildOtherCityObjects(model);
    }

    private void resetFeatureChunks() {
        buildingChunkNr.set(0);
        vegetationChunkNr.set(0);
        transportationChunkNr.set(0);
        tunnelChunkNr.set(0);
        bridgeChunkNr.set(0);
        waterChunkNr.set(0);
        landChunkNr.set(0);
        cityFurnitureChunkNr.set(0);
        otherObjectsChunkNr.set(0);
    }

    private void buildLand(CityDoctorModel model) {
        if (model.getLand().isEmpty()) {
            return;
        }
        TreeView<Renderable> landView = mainWindow.getTerrainView();
        TreeItem<Renderable> landRoot = new TreeItem<>(new AllTerrainNode(model.getLand()));
        landRoot.setExpanded(true);
        landView.setRoot(landRoot);
        buildLandFromList(model.getLand(), landView.getRoot());
        addMoreButtonIfNecessary(model.getLand(), landView, landRoot, landChunkNr);
    }

    private void buildLandFromList(List<CityObject> list, TreeItem<Renderable> root) {
        int landChunk = landChunkNr.get();
        for (int i = landChunk * MAX_FEATURES_PER_CHUNK; i < (landChunk + 1) * MAX_FEATURES_PER_CHUNK
                && i < list.size(); i++) {
            CityObject land = list.get(i);
            Renderable node;
            if (land instanceof ReliefObject relief) {
                node = new ReliefNode(relief);
            } else if (land instanceof TinObject tin) {
                node = new TinNode(tin);
            } else {
                node = new LandUseNode((LandObject) land);
            }
            TreeItem<Renderable> item = new TreeItem<>(node);
            item.setExpanded(true);
            root.getChildren().add(item);
            createGeometryNodes(land, item);
            if (land instanceof ReliefObject relief) {
                createTinNodes(relief, item);
            }
        }
    }

    private void createTinNodes(ReliefObject relief, TreeItem<Renderable> item) {
        if (relief.getComponents().isEmpty()) {
            return;
        }
        AllTinNode allTinText = new AllTinNode(relief.getComponents());
        TreeItem<Renderable> allBpsTextItem = new TreeItem<>(allTinText);
        item.getChildren().add(allBpsTextItem);

        for (TinObject tin : relief.getComponents()) {
            TinNode tinNode = new TinNode(tin);
            TreeItem<Renderable> tinNodeItem = new TreeItem<>(tinNode);
            tinNodeItem.setExpanded(true);
            allBpsTextItem.getChildren().add(tinNodeItem);
            createGeometryNodes(tin, tinNodeItem);
        }
    }

    private void buildCityFurniture(CityDoctorModel model) {
        if (model.getCityFurniture().isEmpty()) {
            return;
        }
        TreeView<Renderable> cityFurnitureView = mainWindow.getCityFurnitureView();
        TreeItem<Renderable> cityFurnitureRoot = new TreeItem<>(new AllCityFurnitureNode(model.getCityFurniture()));
        cityFurnitureRoot.setExpanded(true);
        cityFurnitureView.setRoot(cityFurnitureRoot);
        buildTreeFromList(model.getCityFurniture(), cityFurnitureView.getRoot(), cityFurnitureChunkNr);
        addMoreButtonIfNecessary(model.getCityFurniture(), cityFurnitureView, cityFurnitureRoot, cityFurnitureChunkNr);
    }

    private void buildOtherCityObjects(CityDoctorModel model) {
        if (model.getGenericCityObjects().isEmpty()) {
            return;
        }
        TreeView<Renderable> otherObjectsView = mainWindow.getOtherObjectsView();
        TreeItem<Renderable> otherObjectsRoot = new TreeItem<>(new AllOtherObjectsNode(model.getGenericCityObjects()));
        otherObjectsRoot.setExpanded(true);
        otherObjectsView.setRoot(otherObjectsRoot);
        buildTreeFromList(model.getGenericCityObjects(), otherObjectsView.getRoot(), otherObjectsChunkNr);
        addMoreButtonIfNecessary(model.getGenericCityObjects(), otherObjectsView, otherObjectsRoot, otherObjectsChunkNr);
    }

    private void buildWater(CityDoctorModel model) {
        if (model.getWater().isEmpty()) {
            return;
        }
        TreeView<Renderable> waterView = mainWindow.getWaterView();
        TreeItem<Renderable> waterRoot = new TreeItem<>(new AllWaterNode(model.getWater()));
        waterRoot.setExpanded(true);
        waterView.setRoot(waterRoot);
        buildTreeFromList(model.getWater(), waterView.getRoot(), waterChunkNr);
        addMoreButtonIfNecessary(model.getWater(), waterView, waterRoot, waterChunkNr);
    }

    private void buildBridges(CityDoctorModel model) {
        if (model.getBridges().isEmpty()) {
            return;
        }
        TreeView<Renderable> bridgeView = mainWindow.getBridgeView();
        TreeItem<Renderable> bridgeRoot = new TreeItem<>(new AllBridgesNode(model.getBridges()));
        bridgeRoot.setExpanded(true);
        bridgeView.setRoot(bridgeRoot);
        buildBridgeTreeFromList(model.getBridges(), bridgeView.getRoot());
        addMoreButtonIfNecessary(model.getBridges(), bridgeView, bridgeRoot, bridgeChunkNr);
    }

    private void buildTransportation(List<TransportationObject> trans) {
        if (trans.isEmpty()) {
            return;
        }
        TreeView<Renderable> transView = mainWindow.getTransportationView();
        TreeItem<Renderable> transRoot = new TreeItem<>(new AllTransportationNode(model.getTransportation()));
        transRoot.setExpanded(true);
        transView.setRoot(transRoot);
        buildTreeFromList(trans, transView.getRoot(), transportationChunkNr);
        addMoreButtonIfNecessary(trans, transView, transRoot, transportationChunkNr);
    }

    private void buildTunnel(List<Tunnel> tunnels) {
        if (tunnels.isEmpty()) {
            return;
        }
        TreeView<Renderable> transView = mainWindow.getTunnelView();
        TreeItem<Renderable> transRoot = new TreeItem<>(new AllTunnelsNode(model.getTunnels()));
        transRoot.setExpanded(true);
        transView.setRoot(transRoot);
        buildTreeFromList(tunnels, transView.getRoot(), tunnelChunkNr);
        addMoreButtonIfNecessary(tunnels, transView, transRoot, tunnelChunkNr);
    }

    private void buildVegetation(List<Vegetation> veg) {
        if (veg.isEmpty()) {
            return;
        }
        TreeView<Renderable> vegetationsView = mainWindow.getVegetationView();
        TreeItem<Renderable> vegRoot = new TreeItem<>(new AllVegetationNode(model.getVegetation()));
        vegRoot.setExpanded(true);
        vegetationsView.setRoot(vegRoot);
        buildTreeFromListWithProducer(veg, vegetationsView.getRoot(), vegetationChunkNr, VegetationNode::new);
        addMoreButtonIfNecessary(veg, vegetationsView, vegRoot, vegetationChunkNr);
    }

    private void addMoreButtonIfNecessary(List<? extends CityObject> list, TreeView<Renderable> view,
                                          TreeItem<Renderable> root, AtomicInteger chunkCounter) {
        if ((chunkCounter.get() + 1) * MAX_FEATURES_PER_CHUNK < list.size()) {
            Runnable run = new Runnable() {
                public void run() {
                    // 29 is the height of one item
                    int numberOfDisplayedItems = (int) (view.getHeight() / 29);
                    int selectedIndex = view.getSelectionModel().getSelectedIndex();
                    // remove button node
                    root.getChildren().remove(root.getChildren().size() - 1);
                    chunkCounter.incrementAndGet();
                    int visibleFeatures = countVisibleNodes(view);
                    buildTreeFromList(list, root, chunkCounter);
                    updateTree(root);
                    if (selectedIndex >= 0) {
                        view.getSelectionModel().select(selectedIndex);
                    }
                    if ((chunkCounter.get() + 1) * MAX_FEATURES_PER_CHUNK < list.size()) {
                        root.getChildren().add(new TreeItem<>(new ButtonRenderable(this)));
                    }
                    view.scrollTo(visibleFeatures - numberOfDisplayedItems + 1);
                }
            };
            // add button for expansion
            root.getChildren().add(new TreeItem<>(new ButtonRenderable(run)));
        }
    }

    private void buildBuildings(CityDoctorModel model) {
        if (model.getBuildings().isEmpty()) {
            return;
        }
        TreeView<Renderable> buildingsView = mainWindow.getBuildingsView();
        TreeItem<Renderable> root = new TreeItem<>(new AllBuildingsNode(model.getBuildings()));
        root.setExpanded(true);
        buildBuildingTreeFromList(model.getBuildings(), root);
        addMoreButtonToBuildingsIfNecessary(model.getBuildings(), buildingsView, root, buildingChunkNr);
        buildingsView.setRoot(root);
    }

    private void addMoreButtonToBuildingsIfNecessary(List<Building> list, TreeView<Renderable> view,
                                                     TreeItem<Renderable> root, AtomicInteger chunkCounter) {
        if ((chunkCounter.get() + 1) * MAX_FEATURES_PER_CHUNK < list.size()) {
            Runnable run = new Runnable() {
                public void run() {
                    // 29 is the height of one item
                    int numberOfDisplayedItems = (int) (view.getHeight() / 29);
                    int selectedIndex = view.getSelectionModel().getSelectedIndex();
                    // remove button node
                    root.getChildren().remove(root.getChildren().size() - 1);
                    chunkCounter.getAndIncrement();
                    int visibleFeatures = countVisibleNodes(view);
                    buildBuildingTreeFromList(list, root);
                    updateTree(root);
                    if (selectedIndex >= 0) {
                        view.getSelectionModel().select(selectedIndex);
                    }
                    if ((chunkCounter.get() + 1) * MAX_FEATURES_PER_CHUNK < list.size()) {
                        root.getChildren().add(new TreeItem<>(new ButtonRenderable(this)));
                    }
                    view.scrollTo(visibleFeatures - numberOfDisplayedItems + 1);
                }
            };
            // add button for expansion
            root.getChildren().add(new TreeItem<>(new ButtonRenderable(run)));
        }
    }

    private static int countVisibleNodes(TreeView<Renderable> view) {
        return countVisibleNodes(view.getRoot()) + 1;
    }

    private static int countVisibleNodes(TreeItem<Renderable> root) {
        int visibleFeatures = root.getChildren().size();
        for (TreeItem<Renderable> child : root.getChildren()) {
            if (!child.getChildren().isEmpty() && child.isExpanded()) {
                visibleFeatures += countVisibleNodes(child);
            }
        }
        return visibleFeatures;
    }

    private void buildTreeFromList(List<? extends CityObject> cos, TreeItem<Renderable> root,
                                   AtomicInteger chunkCounter) {
        int chunk = chunkCounter.get();
        for (int i = chunk * MAX_FEATURES_PER_CHUNK; i < (chunk + 1) * MAX_FEATURES_PER_CHUNK && i < cos.size(); i++) {
            CityObject co = cos.get(i);
            CityObjectNode node = new CityObjectNode(co);
            TreeItem<Renderable> nodeItem = new TreeItem<>(node);
            nodeItem.setExpanded(true);
            root.getChildren().add(nodeItem);
            createGeometryNodes(co, nodeItem);
        }
    }

    private <T extends CityObject> void buildTreeFromListWithProducer(List<T> cos, TreeItem<Renderable> root,
                                                                      AtomicInteger chunkCounter, Function<T, Renderable> nodeProducer) {
        int chunk = chunkCounter.get();
        for (int i = chunk * MAX_FEATURES_PER_CHUNK; i < (chunk + 1) * MAX_FEATURES_PER_CHUNK && i < cos.size(); i++) {
            T co = cos.get(i);
            Renderable node = nodeProducer.apply(co);
            TreeItem<Renderable> nodeItem = new TreeItem<>(node);
            nodeItem.setExpanded(true);
            root.getChildren().add(nodeItem);
            createGeometryNodes(co, nodeItem);
        }
    }

    private void clearTrees() {
        mainWindow.getBuildingsView().setRoot(null);
        mainWindow.getVegetationView().setRoot(null);
        mainWindow.getTransportationView().setRoot(null);
        mainWindow.getBridgeView().setRoot(null);
        mainWindow.getWaterView().setRoot(null);
        mainWindow.getTerrainView().setRoot(null);
        mainWindow.getCityFurnitureView().setRoot(null);
        mainWindow.getOtherObjectsView().setRoot(null);
        mainWindow.getErrorTree().getRoot().getChildren().clear();
        mainWindow.getAttributeView().getRoot().getChildren().clear();
        mainWindow.getGlobalErrorsView().getItems().clear();
        clearGeometryTrees();
    }

    private void clearGeometryTrees() {
        highlightController.clearHighlights();
        mainWindow.getPolygonsView().getRoot().getChildren().clear();
        mainWindow.getEdgeView().getRoot().getChildren().clear();
        mainWindow.getVertexView().getRoot().getChildren().clear();
    }

    private void buildBuildingTreeFromList(List<Building> list, TreeItem<Renderable> root) {
        int buildingChunk = buildingChunkNr.get();
        for (int i = buildingChunk * MAX_FEATURES_PER_CHUNK; i < (buildingChunk + 1) * MAX_FEATURES_PER_CHUNK
                && i < list.size(); i++) {
            Building b = list.get(i);
            BuildingNode node = new BuildingNode(b);
            TreeItem<Renderable> item = new TreeItem<>(node);
            item.setExpanded(true);
            root.getChildren().add(item);
            createGeometryNodes(b, item);
            createBoundarySurfaceNodes(b.getBoundarySurfaces(), item);
            createBuildingInstallationNodes(b, item);
            createBuildingRoomNodes(b, item);
            createBuildingFurnitureNodes(b, item);
            createBuildingStoreyNodes(b, item);
            createBuildingUnitNodes(b, item);
            createBuildingPartNodes(b, item);
        }
    }

    private void buildBridgeTreeFromList(List<BridgeObject> list, TreeItem<Renderable> root) {
        int bridgeChunk = bridgeChunkNr.get();
        for (int i = bridgeChunk * MAX_FEATURES_PER_CHUNK; i < (bridgeChunk + 1) * MAX_FEATURES_PER_CHUNK
                && i < list.size(); i++) {
            BridgeObject bridge = list.get(i);
            BridgeNode node = new BridgeNode(bridge);
            TreeItem<Renderable> item = new TreeItem<>(node);
            item.setExpanded(true);
            root.getChildren().add(item);
            createGeometryNodes(bridge, item);
            createBoundarySurfaceNodes(bridge.getBoundarySurfaces(), item);
            createBridgeInstallationNodes(bridge, item);
            createConstructiveElementsNodes(bridge, item);
            createBridgePartNodes(bridge, item);
        }

    }

    private void createBridgeInstallationNodes(BridgeObject bridge, TreeItem<Renderable> root) {
        createInstallationNodes(bridge.getBridgeInstallations(), root);
    }

    private void createConstructiveElementsNodes(BridgeObject bridge, TreeItem<Renderable> item) {
        if (bridge.getConstructiveElements().isEmpty()) {
            return;
        }
        AllBridgeConstructiveElementsNode allBceNode = new AllBridgeConstructiveElementsNode(
                bridge.getConstructiveElements());
        TreeItem<Renderable> allBceNodeTextItem = new TreeItem<>(allBceNode);
        item.getChildren().add(allBceNodeTextItem);

        for (BridgeConstructiveElement bce : bridge.getConstructiveElements()) {
            BridgeConstructiveElementNode bceNode = new BridgeConstructiveElementNode(bce);
            TreeItem<Renderable> bpNodeItem = new TreeItem<>(bceNode);
            bpNodeItem.setExpanded(true);
            allBceNodeTextItem.getChildren().add(bpNodeItem);
            createGeometryNodes(bce, bpNodeItem);
            createBoundarySurfaceNodes(bce.getBoundarySurfaces(), bpNodeItem);
        }
    }

    private void createBridgePartNodes(BridgeObject bridge, TreeItem<Renderable> item) {
        if (bridge.getParts().isEmpty()) {
            return;
        }
        AllBridgePartsNode allBpsText = new AllBridgePartsNode(bridge.getParts());
        TreeItem<Renderable> allBpsTextItem = new TreeItem<>(allBpsText);
        item.getChildren().add(allBpsTextItem);

        for (BridgeObject bp : bridge.getParts()) {
            BridgeNode bpNode = new BridgeNode(bp);
            TreeItem<Renderable> bpNodeItem = new TreeItem<>(bpNode);
            bpNodeItem.setExpanded(true);
            allBpsTextItem.getChildren().add(bpNodeItem);
            createGeometryNodes(bp, bpNodeItem);
            createBoundarySurfaceNodes(bp.getBoundarySurfaces(), bpNodeItem);
        }
    }

    private void createBuildingInstallationNodes(Building ab, TreeItem<Renderable> root) {
        createInstallationNodes(ab.getBuildingInstallations(), root);
    }

    private void createInstallationNodes(List<Installation> installations, TreeItem<Renderable> root) {
        if (installations.isEmpty()) {
            return;
        }

        AllInstallationsNode allBiNode = new AllInstallationsNode(installations);
        TreeItem<Renderable> allBpsTextItem = new TreeItem<>(allBiNode);
        root.getChildren().add(allBpsTextItem);
        for (Installation bi : installations) {
            InstallationNode biNode = new InstallationNode(bi);
            TreeItem<Renderable> biItem = new TreeItem<>(biNode);
            biItem.setExpanded(true);
            allBpsTextItem.getChildren().add(biItem);
            createGeometryNodes(bi, biItem);
            createBoundarySurfaceNodes(bi.getBoundarySurfaces(), biItem);
        }
    }

    private void createBuildingRoomNodes(Building ab, TreeItem<Renderable> root) {
        createRoomNodes(ab.getBuildingRooms(), root);
    }

    private void createRoomNodes(List<? extends AbstractRoom> rooms, TreeItem<Renderable> root) {
        if (rooms.isEmpty()) {
            return;
        }

        AllRoomsNode allRoNode = new AllRoomsNode(rooms);
        TreeItem<Renderable> allRoNodeTextItem = new TreeItem<>(allRoNode);
        root.getChildren().add(allRoNodeTextItem);
        for (AbstractRoom room : rooms) {
            RoomNode roomNode = new RoomNode(room);
            TreeItem<Renderable> roomNodeTextItem = new TreeItem<>(roomNode);
            roomNodeTextItem.setExpanded(true);
            allRoNodeTextItem.getChildren().add(roomNodeTextItem);
            createGeometryNodes(room, roomNodeTextItem);
            createBoundarySurfaceNodes(room.getBoundarySurfaces(), roomNodeTextItem);
        }

    }

    private void createBuildingStoreyNodes(Building ab, TreeItem<Renderable> root) {
        List<Storey> storeys = ab.getBuildingStoreys();
        if (storeys.isEmpty()) {
            return;
        }
        AllStoreysNode allSNode = new AllStoreysNode(storeys);
        TreeItem<Renderable> allSNodeTextItem = new TreeItem<>(allSNode);
        root.getChildren().add(allSNodeTextItem);
        for (Storey storey : storeys) {
            StoreyNode storeyNode = new StoreyNode(storey);
            TreeItem<Renderable> storeyNodeTextItem = new TreeItem<>(storeyNode);
            storeyNodeTextItem.setExpanded(true);
            allSNodeTextItem.getChildren().add(storeyNodeTextItem);
            createGeometryNodes(storey, storeyNodeTextItem);
            createBoundarySurfaceNodes(storey.getBoundarySurfaces(), storeyNodeTextItem);
        }
    }


    public void createBuildingUnitNodes(Building ab, TreeItem<Renderable> root) {
        List<BuildingUnit> buildingUnits = ab.getBuildingUnits();
        if (buildingUnits.isEmpty()) {
            return;
        }
        AllBuildingUnitsNode allSNode = new AllBuildingUnitsNode(buildingUnits);
        TreeItem<Renderable> allSNodeTextItem = new TreeItem<>(allSNode);
        root.getChildren().add(allSNodeTextItem);
        for (BuildingUnit buildingUnit : buildingUnits) {
            BuildingUnitNode buildingUnitNode = new BuildingUnitNode(buildingUnit);
            TreeItem<Renderable> storeyNodeTextItem = new TreeItem<>(buildingUnitNode);
            storeyNodeTextItem.setExpanded(true);
            allSNodeTextItem.getChildren().add(storeyNodeTextItem);
            createGeometryNodes(buildingUnit, storeyNodeTextItem);
            createBoundarySurfaceNodes(buildingUnit.getBoundarySurfaces(), storeyNodeTextItem);
        }
    }

    private void createBuildingFurnitureNodes(Building ab, TreeItem<Renderable> root) {
        createFurnitureNodes(ab.getBuildingRoomFurnitureList(), root);
    }

    private void createFurnitureNodes(List<? extends AbstractFurniture> furniture, TreeItem<Renderable> root) {
        if (furniture.isEmpty()) {
            return;
        }

        AllFurnitureNode allFnNode = new AllFurnitureNode(furniture);
        TreeItem<Renderable> allFnNodeTextItem = new TreeItem<>(allFnNode);
        root.getChildren().add(allFnNodeTextItem);
        for (AbstractFurniture fn : furniture) {
            FurnitureNode fnNode = new FurnitureNode(fn);
            TreeItem<Renderable> fnNodeTextItem = new TreeItem<>(fnNode);
            fnNodeTextItem.setExpanded(true);
            allFnNodeTextItem.getChildren().add(fnNodeTextItem);
            createGeometryNodes(fn, fnNodeTextItem);
            createBoundarySurfaceNodes(fn.getBoundarySurfaceList(), fnNodeTextItem);
        }
    }

    private void createBoundarySurfaceNodes(List<BoundarySurface> bsList, TreeItem<Renderable> root) {
        if (bsList.isEmpty()) {
            return;
        }
        AllBoundarySurfacesNode allBsText = new AllBoundarySurfacesNode(bsList);
        TreeItem<Renderable> allBpsTextItem = new TreeItem<>(allBsText);
        root.getChildren().add(allBpsTextItem);

        allBpsTextItem.getChildren().add(new TreeItem<>());

        allBpsTextItem.expandedProperty().addListener((obs, old, newV) -> {
            if (Boolean.TRUE.equals(newV) && allBpsTextItem.getChildren().size() == 1) {
                Platform.runLater(() -> {
                    allBpsTextItem.getChildren().clear();
                    for (BoundarySurface bs : bsList) {
                        BoundarySurfaceNode bsText = new BoundarySurfaceNode(bs);
                        TreeItem<Renderable> bsTextItem = new TreeItem<>(bsText);
                        bsTextItem.setExpanded(true);
                        allBpsTextItem.getChildren().add(bsTextItem);
                        createGeometryNodes(bs, bsTextItem);
                        createOpeningNodes(bs.getOpenings(), bsTextItem);
                    }
                    updateTree(allBpsTextItem);
                });
            }
        });
    }

    private void createOpeningNodes(List<Opening> openings, TreeItem<Renderable> root) {
        if (openings == null || openings.isEmpty()) {
            return;
        }
        AllOpeningsNode allOpeningsNode = new AllOpeningsNode(openings);
        TreeItem<Renderable> allOpeningsItem = new TreeItem<>(allOpeningsNode);
        for (Opening o : openings) {
            OpeningNode openingNode = new OpeningNode(o);
            TreeItem<Renderable> openingsItem = new TreeItem<>(openingNode);
            allOpeningsItem.getChildren().add(openingsItem);
        }
        root.getChildren().add(allOpeningsItem);
    }

    private void createBuildingPartNodes(Building ab, TreeItem<Renderable> item) {
        if (ab.getBuildingParts().isEmpty()) {
            return;
        }
        AllBuildingPartsNode allBpsText = new AllBuildingPartsNode(ab.getBuildingParts());
        TreeItem<Renderable> allBpsTextItem = new TreeItem<>(allBpsText);
        item.getChildren().add(allBpsTextItem);

        for (BuildingPart bp : ab.getBuildingParts()) {
            BuildingPartNode bpNode = new BuildingPartNode(bp);
            TreeItem<Renderable> bpNodeItem = new TreeItem<>(bpNode);
            bpNodeItem.setExpanded(true);
            allBpsTextItem.getChildren().add(bpNodeItem);
            createGeometryNodes(bp, bpNodeItem);
            createBoundarySurfaceNodes(bp.getBoundarySurfaces(), bpNodeItem);
        }
    }

    private void createGeometryNodes(CityObject co, TreeItem<Renderable> item) {
        for (Geometry geom : co.getGeometries()) {
            GeometryNode geomNode = new GeometryNode(geom);
            TreeItem<Renderable> geomItem = new TreeItem<>(geomNode);
            item.getChildren().add(geomItem);
        }
    }

    public void startChecks(ValidationConfiguration config, ProgressListener l) {
        if (model == null) {
            logger.warn(Localization.getText("CityDoctorController.noDatamodel"));
            return;
        }
        if (sourceFile == null) {
            logger.warn(Localization.getText("CityDoctorController.noSourceFile"));
            return;
        }
        try {
            Platform.runLater(() -> {
                mainWindow.getCheckButton().setDisable(true);
                mainWindow.getErrorTree().getRoot().getChildren().clear();
                mainWindow.getGlobalErrorsView().getItems().clear();
            });
            config.setNumberOfRoundingPlacesInGlobalParameters(currentConfig.getNumberOfRoundingPlaces());
            config.setParserConfig(currentConfig);
            currentChecker = new Checker(config, model);
            currentChecker.runChecks(l);

            Platform.runLater(() -> {
                // apply check results to tree views
                updateFeatureTrees();
                updateTree(mainWindow.getPolygonsView().getRoot());
                for (CheckError e : model.getGlobalErrors()) {
                    if (e instanceof SchematronError se) {
                        mainWindow.getGlobalErrorsView().getItems().add(se.getErrorIdString());
                    }
                }
                renderer.refresh();
                mainWindow.getWriteReportButton().setDisable(false);
            });
        } finally {
            Platform.runLater(() -> mainWindow.getCheckButton().setDisable(false));
        }
    }

    void updateFeatureTrees() {
        updateTree(mainWindow.getBuildingsView().getRoot());
        updateTree(mainWindow.getVegetationView().getRoot());
        updateTree(mainWindow.getBridgeView().getRoot());
        updateTree(mainWindow.getTerrainView().getRoot());
        updateTree(mainWindow.getCityFurnitureView().getRoot());
        updateTree(mainWindow.getTransportationView().getRoot());
        updateTree(mainWindow.getWaterView().getRoot());
        updateTree(mainWindow.getOtherObjectsView().getRoot());
        renderer.updateErrors();
    }

    private void updateTree(TreeItem<Renderable> root) {
        if (root == null) {
            return;
        }
        for (TreeItem<Renderable> item : root.getChildren()) {
            updateItem(item);
        }
    }

    private void updateItem(TreeItem<Renderable> item) {
        if (item.getValue() == null) {
            return;
        }
        item.getValue().refreshTextColor();
        for (TreeItem<Renderable> child : item.getChildren()) {
            updateItem(child);
        }
    }

    public void writePdfReport(File pdfFile) {
        if (currentChecker == null) {
            return;
        }
        currentChecker.writePdfReport(pdfFile.getAbsolutePath());
    }

    public void writeXmlReport(File xmlFile) {
        if (currentChecker == null) {
            return;
        }
        currentChecker.writeXmlReport(xmlFile.getAbsolutePath());
    }

    private Map<ErrorId, ErrorStat> getErrorStats() {
        Map<ErrorId, ErrorStat> stats = new HashMap<>();
        addErrorStats(stats, model.getBridges());
        addErrorStats(stats, model.getBuildings());
        addErrorStats(stats, model.getLand());
        addErrorStats(stats, model.getTransportation());
        addErrorStats(stats, model.getVegetation());
        addErrorStats(stats, model.getWater());
        addErrorStats(stats, model.getCityFurniture());
        addErrorStats(stats, model.getGenericCityObjects());
        return stats;
    }

    private void addErrorStats(Map<ErrorId, ErrorStat> stats, List<? extends CityObject> cos) {
        List<CheckError> errors = new ArrayList<>();
        for (CityObject co : cos) {
            co.collectContainedErrors(errors);
        }
        Set<CheckError> filteredErrors = new HashSet<>(errors);
        for (CheckError error : filteredErrors) {
            ErrorStat stat = stats.computeIfAbsent(error.getErrorId(), ErrorStat::new);
            stat.incrementCount();
        }
    }

    public void export(CityObject co) {
        if (model == null) {
            return;
        }
        CityDoctorModel newModel = new CityDoctorModel(model.getParserConfig(), model.getFile());
        newModel.setCityModel(new CityModel());
        newModel.addCityObject(co);
        newModel.setParsedCityGMLVersion(model.getCityGMLVersion());

        FileChooser fc = new FileChooser();
        fc.getExtensionFilters().add(new ExtensionFilter("CityGML", "*.gml"));
        fc.getExtensionFilters().add(new ExtensionFilter("OFF - Object File Format", "*.off"));
        File dir = new File(Settings.get(Settings.LAST_OPEN_FOLDER, ""));
        if (dir.exists() && dir.isDirectory()) {
            fc.setInitialDirectory(dir);
        } else {
            Settings.set(Settings.LAST_OPEN_FOLDER, "");
        }
        File f = fc.showSaveDialog(mainWindow.getMainStage());
        if (f != null) {
            Settings.set(Settings.LAST_OPEN_FOLDER, f.getParent());
            try {
                newModel.saveAs(f.getAbsolutePath(), true);
            } catch (CityDoctorWriteException e) {
                logger.error(e);
            }
        }
    }

    public void showWorld() {
        if (model != null) {
            renderer.render(model);
        }
    }

    public void storeModel(File f, boolean saveQualityAde) throws CityDoctorWriteException {
        if (model == null) {
            return;
        }
        model.saveAs(f.getAbsolutePath(), saveQualityAde);
    }

    public void searchFeature(String searchString, FeatureType selectedTab) {
        if (model == null || searchString == null || searchString.trim().isEmpty()) {
            return;
        }
        mainWindow.unselectEverything();
        resetFeatureChunks();
        if (selectedTab == FeatureType.BUILDING) {
            List<Building> foundBuildings = filterFeatures(searchString, model.getBuildings());
            TreeView<Renderable> buildingsView = mainWindow.getBuildingsView();
            TreeItem<Renderable> root = buildingsView.getRoot();
            root.getChildren().clear();
            buildBuildingTreeFromList(foundBuildings, root);
            updateTree(root);
            addMoreButtonToBuildingsIfNecessary(foundBuildings, buildingsView, root, buildingChunkNr);
        } else {
            TreeView<Renderable> view;
            List<? extends CityObject> cos;
            AtomicInteger chunkCounter;
            switch (selectedTab) {
                case VEGETATION:
                    view = mainWindow.getVegetationView();
                    cos = filterFeatures(searchString, model.getVegetation());
                    chunkCounter = vegetationChunkNr;
                    break;
                case BRIDGE:
                    view = mainWindow.getBridgeView();
                    cos = filterFeatures(searchString, model.getBridges());
                    chunkCounter = bridgeChunkNr;
                    break;
                case TRANSPORTATION:
                    view = mainWindow.getTransportationView();
                    cos = filterFeatures(searchString, model.getTransportation());
                    chunkCounter = transportationChunkNr;
                    break;
                case TUNNEL:
                    view = mainWindow.getTunnelView();
                    cos = model.getTunnels();
                    chunkCounter = tunnelChunkNr;
                    break;
                case WATER:
                    view = mainWindow.getWaterView();
                    cos = filterFeatures(searchString, model.getWater());
                    chunkCounter = waterChunkNr;
                    break;
                case LAND:
                    view = mainWindow.getTerrainView();
                    cos = filterFeatures(searchString, model.getLand());
                    chunkCounter = landChunkNr;
                    break;
                case CITY_FURNITURE:
                    view = mainWindow.getCityFurnitureView();
                    cos = filterFeatures(searchString, model.getCityFurniture());
                    chunkCounter = cityFurnitureChunkNr;
                    break;
                case GENERIC_CITY_OBJECT:
                    view = mainWindow.getOtherObjectsView();
                    cos = filterFeatures(searchString, model.getGenericCityObjects());
                    chunkCounter = otherObjectsChunkNr;
                    break;
                default:
                    throw new IllegalStateException("Unknown selected feature tab");
            }
            TreeItem<Renderable> root = view.getRoot();
            root.getChildren().clear();
            buildTreeFromList(cos, root, chunkCounter);
            updateTree(root);
            addMoreButtonIfNecessary(cos, view, root, chunkCounter);
        }
    }

    private <T extends CityObject> List<T> filterFeatures(String searchString, List<T> features) {
        List<T> foundFeatures = new ArrayList<>();
        for (T t : features) {
            if (t.getGmlId().getGmlString().contains(searchString)) {
                foundFeatures.add(t);
            }
        }
        return foundFeatures;
    }

    public void resetSearch(FeatureType selectedTab) {
        if (model == null) {
            return;
        }
        resetFeatureChunks();
        if (selectedTab == FeatureType.BUILDING) {
            // buildings are handled differently, because of surface and building
            // installations
            TreeItem<Renderable> root = mainWindow.getBuildingsView().getRoot();
            if (root != null) {
                root.getChildren().clear();
                buildBuildingTreeFromList(model.getBuildings(), root);
                updateTree(root);
                addMoreButtonToBuildingsIfNecessary(model.getBuildings(), mainWindow.getBuildingsView(), root,
                        buildingChunkNr);
            }
        } else {
            TreeView<Renderable> view;
            List<? extends CityObject> cos;
            AtomicInteger chunkCounter;
            switch (selectedTab) {
                case VEGETATION:
                    view = mainWindow.getVegetationView();
                    cos = model.getVegetation();
                    chunkCounter = vegetationChunkNr;
                    break;
                case BRIDGE:
                    view = mainWindow.getBridgeView();
                    cos = model.getBridges();
                    chunkCounter = bridgeChunkNr;
                    break;
                case TRANSPORTATION:
                    view = mainWindow.getTransportationView();
                    cos = model.getTransportation();
                    chunkCounter = transportationChunkNr;
                    break;
                case TUNNEL:
                    view = mainWindow.getTunnelView();
                    cos = model.getTunnels();
                    chunkCounter = tunnelChunkNr;
                    break;
                case WATER:
                    view = mainWindow.getWaterView();
                    cos = model.getWater();
                    chunkCounter = waterChunkNr;
                    break;
                case LAND:
                    view = mainWindow.getTerrainView();
                    cos = model.getLand();
                    chunkCounter = landChunkNr;
                    break;
                case CITY_FURNITURE:
                    view = mainWindow.getCityFurnitureView();
                    cos = model.getCityFurniture();
                    chunkCounter = cityFurnitureChunkNr;
                    break;
                case GENERIC_CITY_OBJECT:
                    view = mainWindow.getOtherObjectsView();
                    cos = model.getGenericCityObjects();
                    chunkCounter = otherObjectsChunkNr;
                    break;
                default:
                    throw new IllegalStateException("Unknown selected feature tab");
            }
            TreeItem<Renderable> root = view.getRoot();
            root.getChildren().clear();
            buildTreeFromList(cos, root, chunkCounter);
            updateTree(root);
            addMoreButtonIfNecessary(cos, view, root, chunkCounter);
        }
    }

    public Series<String, Number> createErrorSeries() {
        Series<String, Number> series = new Series<>();
        if (model == null) {
            return series;
        }
        Map<ErrorId, ErrorStat> errorStats = getErrorStats();
        for (ErrorStat es : errorStats.values()) {
            series.getData().add(new Data<>(es.getErrorId().toString(), es.getCount()));
        }
        return series;
    }

    public String getFileName() {
        return model.getFileName();
    }

    public void fillTreeViewWithErrorBuildings() {
        TreeView<Renderable> buildingsView = mainWindow.getBuildingsView();
        TreeItem<Renderable> root = buildingsView.getRoot();
        if (model == null) {
            return;
        }
        if (root == null) {
            return;
        }
        mainWindow.resetSearchBar();
        root.getChildren().clear();
        List<Building> errorBuildings = new ArrayList<>();
        for (Building b : model.getBuildings()) {
            if (b.containsAnyError()) {
                errorBuildings.add(b);
            }
        }
        resetFeatureChunks();
        buildBuildingTreeFromList(errorBuildings, root);
        updateTree(root);
        addMoreButtonToBuildingsIfNecessary(errorBuildings, buildingsView, root, buildingChunkNr);
    }

    public void fillTreeViewWithErrorVegetation() {
        if (model == null) {
            return;
        }
        fillTreeViewWithErrorCityObjects(mainWindow.getVegetationView(), model.getVegetation(), vegetationChunkNr);
    }

    public void fillTreeViewWithErrorBridges() {
        if (model == null) {
            return;
        }
        fillTreeViewWithErrorCityObjects(mainWindow.getBridgeView(), model.getBridges(), bridgeChunkNr);
    }

    public void fillTreeViewWithErrorLand() {
        if (model == null) {
            return;
        }
        fillTreeViewWithErrorCityObjects(mainWindow.getTerrainView(), model.getLand(), landChunkNr);
    }

    public void fillTreeViewWithErrorCityFurniture() {
        if (model == null) {
            return;
        }
        fillTreeViewWithErrorCityObjects(mainWindow.getCityFurnitureView(),
                model.getCityFurniture(), cityFurnitureChunkNr);
    }

    public void fillTreeViewWithErrorOtherObjects() {
        if (model == null) {
            return;
        }
        fillTreeViewWithErrorCityObjects(mainWindow.getOtherObjectsView(),
                model.getGenericCityObjects(), otherObjectsChunkNr);
    }

    public void fillTreeViewWithErrorTransportation() {
        if (model == null) {
            return;
        }
        fillTreeViewWithErrorCityObjects(mainWindow.getTransportationView(), model.getTransportation(),
                transportationChunkNr);
    }

    public void fillTreeViewWithErrorTunnel() {
        if (model == null) {
            return;
        }
        fillTreeViewWithErrorCityObjects(mainWindow.getTunnelView(), model.getTunnels(), tunnelChunkNr);
    }

    public void fillTreeViewWithErrorWater() {
        if (model == null) {
            return;
        }
        fillTreeViewWithErrorCityObjects(mainWindow.getWaterView(), model.getWater(), waterChunkNr);
    }

    private void fillTreeViewWithErrorCityObjects(TreeView<Renderable> treeView, List<? extends CityObject> objects,
                                                  AtomicInteger chunkCounter) {
        TreeItem<Renderable> root = treeView.getRoot();
        if (root == null) {
            return;
        }
        mainWindow.resetSearchBar();
        root.getChildren().clear();
        List<CityObject> errorObjects = new ArrayList<>();
        for (CityObject co : objects) {
            if (co.containsAnyError()) {
                errorObjects.add(co);
            }
        }
        buildTreeFromList(errorObjects, root, chunkCounter);
        updateTree(root);
        addMoreButtonIfNecessary(errorObjects, treeView, root, chunkCounter);
    }

    public void fillTreeViewFromFeatures(FeatureType selectedTab) {
        if (model == null) {
            return;
        }
        mainWindow.resetSearchBar();
        switch (selectedTab) {
            case BUILDING:
                buildBuildings(model);
                updateTree(mainWindow.getBuildingsView().getRoot());
                break;
            case VEGETATION:
                buildVegetation(model.getVegetation());
                updateTree(mainWindow.getVegetationView().getRoot());
                break;
            case BRIDGE:
                buildBridges(model);
                updateTree(mainWindow.getBridgeView().getRoot());
                break;
            case LAND:
                buildLand(model);
                updateTree(mainWindow.getTerrainView().getRoot());
                break;
            case TRANSPORTATION:
                buildTransportation(model.getTransportation());
                updateTree(mainWindow.getTransportationView().getRoot());
                break;
            case WATER:
                buildWater(model);
                updateTree(mainWindow.getWaterView().getRoot());
                break;
            case CITY_FURNITURE:
                buildCityFurniture(model);
                updateTree(mainWindow.getCityFurnitureView().getRoot());
                break;
            case GENERIC_CITY_OBJECT:
                buildOtherCityObjects(model);
                updateTree(mainWindow.getOtherObjectsView().getRoot());
                break;
            default:
                throw new IllegalStateException();
        }
    }

    public CityDoctorModel getModel() {
        return model;
    }

    public void showView(View v) {
        v.fireOnShowEvent(model, currentChecker, mainWindow);
    }

    public void askAndSave() {
        boolean saveWithQualityAde = true;
        if (model.isValidated()) {
            ButtonType yesBtn = ButtonType.YES;
            ButtonType noBtn = ButtonType.NO;
            Dialog<ButtonType> dialog = new Dialog<>();
            dialog.setTitle("Save with QualityADE?");
            dialog.setContentText("Save with QualityADE information?");
            dialog.getDialogPane().getButtonTypes().add(yesBtn);
            dialog.getDialogPane().getButtonTypes().add(noBtn);

            Optional<ButtonType> result = dialog.showAndWait();
            if (result.isPresent() && result.get() == ButtonType.NO) {
                saveWithQualityAde = false;
            }
        }

        FileChooser fc = new FileChooser();
        fc.getExtensionFilters().add(new ExtensionFilter("CityGML", "*.gml"));
        fc.getExtensionFilters().add(new ExtensionFilter("OFF - Object File Format", "*.off"));
        File dir = new File(Settings.get(Settings.LAST_OPEN_FOLDER, ""));
        if (dir.exists() && dir.isDirectory()) {
            fc.setInitialDirectory(dir);
        } else {
            Settings.set(Settings.LAST_OPEN_FOLDER, "");
        }
        File f = fc.showSaveDialog(mainWindow.getMainStage());
        if (f != null) {
            Settings.set(Settings.LAST_OPEN_FOLDER, f.getParent());
            try {
                storeModel(f, saveWithQualityAde);
                logger.info("Saving successful");
            } catch (CityDoctorWriteException e) {
                mainWindow.showExceptionDialog(e);
            }
        }
    }

    public void delete(TreeItem<Renderable> selectedItem) {
        if (model == null) {
            return;
        }
        Renderable render = selectedItem.getValue();
        if (render instanceof BuildingNode node) {
            model.getBuildings().remove(node.getBuilding());
            mainWindow.getBuildingsView().getRoot().getChildren().remove(selectedItem);
        }

    }

    public void errorFilterIndexChanged(Number newV) {
        mainWindow.getMeshGroup().getChildren().clear();
        mainWindow.unselectEverything();
        if (newV.intValue() == 0) {
            fillTreeViewFromFeatures(mainWindow.getSelectedTab());
        } else if (newV.intValue() == 1) {
            switch (mainWindow.getSelectedTab()) {
                case BUILDING:
                    fillTreeViewWithErrorBuildings();
                    break;
                case VEGETATION:
                    fillTreeViewWithErrorVegetation();
                    break;
                case BRIDGE:
                    fillTreeViewWithErrorBridges();
                    break;
                case LAND:
                    fillTreeViewWithErrorLand();
                    break;
                case TRANSPORTATION:
                    fillTreeViewWithErrorTransportation();
                    break;
                case WATER:
                    fillTreeViewWithErrorWater();
                    break;
                case CITY_FURNITURE:
                    fillTreeViewWithErrorCityFurniture();
                    break;
                case GENERIC_CITY_OBJECT:
                    fillTreeViewWithErrorOtherObjects();
                    break;
                default:
                    throw new IllegalStateException("Unknown selected feature tab: " + mainWindow.getSelectedTab());
            }
        } else {
            throw new IllegalStateException("Unknown filter index selected: " + newV);
        }
    }

    public ParserConfiguration getCurrentConfig() {
        return currentConfig;
    }

    public void setOriginBB(BoundingBox origin) {
        this.originBB = origin;
    }

    public BoundingBox getOriginBB() {
        return originBB;
    }

}
